import json


class Request:
    def __init__(self, scope, receive, send):
        scope['request'] = self
        self._scope = scope
        self._receive = receive
        self._send = send

    async def body(self):
        chunks = []
        while True:
            receive = await self._receive()
            chunks.append(receive['body'])
            if not receive.get('more_body'):
                break

        return b''.join(chunks)

    @property
    def headers(self):
        if hasattr(self, '_headers'):
            return self._headers
        else:
            self._headers = {
                h[0].decode(): h[1].decode() for h in self._scope['headers']
            }
            return self._headers

    async def json(self):
        if hasattr(self, '_json'):
            return self._json
        else:
            self._json = json.loads(await self.body())  # TODO: in executor
            return self._json

    @property
    def path_params(self):
        return self._scope.get('path_params', {})

    @property
    def query(self):
        # TODO: MultiDict
        query_string = self._scope.get('query_string').decode()
        return {
            q.split('=')[0]: q.split('=')[1] for q in query_string.split('&')
        }



from functools import partial

from collections import defaultdict
from .responses import Response
from .request import Request


class Resolver:
    def __init__(self):
        self.routes = defaultdict(list)

    async def __call__(self, scope, receive, send):
        scope['request'] = request = Request(scope, receive, send)

        request.app = app = scope.get('app')
        request.path = path = scope['path']
        request.method = method = scope['method']

        for route in self.routes.get(method, []) + self.routes.get('*', []):
            if route.path == path:
                handler = route.handler
                for middleware in app.middlewares[::-1] + route.middlewares:
                    if middleware not in route.ignore_middlewares:
                        handler = partial(middleware, handler)

                response = await handler(request)

                break
        else:
            response = Response('<h1>404 Not found</h1>', 404)

        await response(scope, receive, send)

    def add_route(self, route):
        for method in route.methods:
            self.routes[method].append(route)



from .resolver import Resolver
from .routes import Route
from .utils import run_func, prettyformat


class App:
    def __init__(self, debug=False, startup=[], shutdown=[], middlewares=[]):
        self.debug = debug
        self.http = Resolver()
        self.startup = startup
        self.shutdown = shutdown
        self.middlewares = middlewares

    async def __call__(self, scope, receive, send):
        print(prettyformat(scope)) if self.debug else None

        # TODO: ASGI middlewares
        scope['app'] = self

        if scope['type'] == 'lifespan':
            await self.lifespan(scope, receive, send)
        elif scope['type'] == 'http':
            await self.http(scope, receive, send)

    async def lifespan(self, scope, receive, send):
        while True:
            message = await receive()
            if message['type'] == 'lifespan.startup':
                for func in self.startup:
                    await run_func(func, self)
                await send({'type': 'lifespan.startup.complete'})

            elif message['type'] == 'lifespan.shutdown':
                for func in self.shutdown:
                    await run_func(func, self)
                await send({'type': 'lifespan.shutdown.complete'})
                return

    def route(self, *args, **kwargs):
        self.http.add_route(Route(*args, **kwargs))



import asyncio
import pprint
from functools import partial

prettyformat = pprint.PrettyPrinter(indent=2).pformat


async def run_in_executor(func, *args, **kwargs):
    if kwargs:
        func = partial(func, **kwargs)

    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, func, *args)


async def run_func(func, *args, **kwargs):
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)

    return await run_in_executor(func, *args, **kwargs)



import re


class Route:
    def __init__(
        self,
        path,
        handler,
        methods=['GET'],
        middlewares=[],
        ignore_middlewares=[],
    ):
        self.path = path
        self.methods = methods
        self.handler = handler
        self.middlewares = middlewares
        self.ignore_middlewares = ignore_middlewares

        groups = re.findall(r'<(\w+):(\w+)>', self.path)
        if groups:
            self.path_params_cast = groups



from distutils.core import setup
setup(
  name='Ancalagon',
  packages=['Ancalagon'],
  version='0.0.1a',
  license='lgpl-3.0',
  description='ASGI Web framework',
  author='Nimond',
  author_email='dndnomin@gmail.com',
  url='https://github.com/Nimond/Ancalagon',
  download_url='https://github.com/Nimond/Ancalagon/archive/0.0.1a.tar.gz',
  keywords=['ASGI', 'WEB', 'FRAMEWORK', 'BLACK'],
  install_requires=[],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)
